@@ -2,8 +2,9 @@ import serial
 import threading
 import time
 import atexit
+import json
 
-print("Electron source controller: 0.0.16")
+print("Electron source controller: 0.0.17")
 
 from collections import deque
 
@@ -138,7 +139,7 @@ class ElectronGunControl:
     POLARITY_POS = b'p'
     POLARITY_NEG = b'n'
 
-    def __init__(self, portFile = None):
+    def __init__(self, portFile = None, commandRetries = 3):
         if not portFile:
             portFile = '/dev/ttyU0'
             try:
@@ -151,6 +152,8 @@ class ElectronGunControl:
 
         self.bufferInput = ElectronGunRingbuffer()
 
+        self._commandRetries = commandRetries
+
         self.port = False
         self.portFile = portFile
         self.port = serial.Serial(portFile, baudrate=19200, bytesize=serial.EIGHTBITS, parity=serial.PARITY_NONE, stopbits=serial.STOPBITS_ONE, timeout=30)
@@ -219,9 +222,28 @@ class ElectronGunControl:
         self.messageFilter = filter
 
         # Now wait till we get a response from our processing thread ...
+        remainingRetries = int(self._commandRetries)
+        if remainingRetries is None:
+            remainingRetries = 0
+        elif remainingRetries < 0:
+            remainingRetries = 0
+
         while self.messageResponse == None:
-            self.messageConditionVariable.wait()
+            if not self.messageConditionVariable.wait(timeout = 10):
+                print(f"Timeout while waiting for {self.messageFilter} (Retries remaining: {remainingRetries})")
+                if (remainingRetries > 0) and (self._lastcommand is not None):
+                    # Retry command ...
+                    self.port.write(self._lastcommand)
+                    remainingRetries = remainingRetries - 1
+                else:
+                    self.jabber(sync = False)
+                    self.messageFilter = None
+                    self.messageResponse = None
+                    self._lastcommand = None
+                    self.messageConditionVariable.release()
+                    return None
         # Reset message filter, copy and release response ...
+        self._lastcommand = None
         self.messageFilter = None
         retval = self.messageResponse
         self.messageResponse = None
@@ -434,6 +456,7 @@ class ElectronGunControl:
         if self.port == False:
             raise ElectronGunNotConnected("Electron gun currently not connected")
         self.port.write(b'$$$id\n')
+        self._lastcommand = b'$$$id\n'
         if sync:
             return self.internal__waitForMessageFilter("id")
         else:
@@ -452,12 +475,11 @@ class ElectronGunControl:
         elif channel == 4:
             cmd = b'$$$psugetv4\n'
         else:
-            raiseElectronGunInvalidParameterException("Power supply channel has to be in range 1 to 4")
+            raise ElectronGunInvalidParameterException("Power supply channel has to be in range 1 to 4")
         self.port.write(cmd)
+        self._lastcommand = cmd
         if sync:
             res = self.internal__waitForMessageFilter("v{}".format(channel))
-            if self.stabilizationDelay:
-                time.sleep(self.stabilizationDelay)
             return res
         else:
             return None
@@ -475,7 +497,7 @@ class ElectronGunControl:
         currentWhenelt = self.getPSUCurrent(2, sync = True)
         currentFocus = self.getPSUCurrent(3, sync = True)
 
-        return currentCathode + currentWhenelt + currentFocus
+        return currentCathode - currentWhenelt - currentFocus
 
     def getPSUCurrent(self, channel, *ignore, sync = False):
         if self.port == False:
@@ -492,11 +514,10 @@ class ElectronGunControl:
         else:
             raiseElectronGunInvalidParameterException("Power supply channel has to be in range 1 to 4")
         self.port.write(cmd)
+        self._lastcommand = cmd
 
         if sync:
             res = self.internal__waitForMessageFilter("a{}".format(channel))
-            if self.stabilizationDelay:
-                time.sleep(self.stabilizationDelay)
             return res
         else:
             return None
@@ -506,6 +527,7 @@ class ElectronGunControl:
             raise ElectronGunNotConnected("Electron gun currently not connected")
         cmd = b'$$$psumode\n'
         self.port.write(cmd)
+        self._lastcommand = cmd
         if sync:
             return self.internal__waitForMessageFilter("psustate")
         else:
@@ -516,6 +538,7 @@ class ElectronGunControl:
             raise ElectronGunNotConnected("Electron gun currently not connected")
         cmd = b'$$$fila\n'
         self.port.write(cmd)
+        self._lastcommand = cmd
         if sync:
             return self.internal__waitForMessageFilter("af")
         else:
@@ -526,6 +549,7 @@ class ElectronGunControl:
             raise ElectronGunNotConnected("Electron gun currently not connected")
         cmd = b'$$$off\n'
         self.port.write(cmd)
+        self._lastcommand = cmd
         # Here we add a small delay to allow the serial buffer to be
         # fully flushed before we terminate our process so we can make sure
         # we really transmit the off condition
@@ -536,6 +560,7 @@ class ElectronGunControl:
             raise ElectronGunNotConnected("Electron gun currently not connected")
         cmd = b'$$$noprotection\n'
         self.port.write(cmd)
+        self._lastcommand = cmd
 
     def setPSUPolarity(self, channel, polarity, *ignore, sync = False):
         if self.port == False:
@@ -557,6 +582,7 @@ class ElectronGunControl:
 
         cmd = cmd + polarity + b'\n'
         self.port.write(cmd)
+        self._lastcommand = cmd
 
         if sync:
             time.sleep(20)
@@ -576,6 +602,7 @@ class ElectronGunControl:
         else:
             raise ElectronGunInvalidParameterException("Power supply channel has to be in range 1 to 4")
         self.port.write(cmd)
+        self._lastcommand = cmd
 
         if self.stabilizationDelay and sync:
             time.sleep(self.stabilizationDelay)
@@ -595,6 +622,7 @@ class ElectronGunControl:
         else:
             raiseElectronGunInvalidParameterException("Power supply channel has to be in range 1 to 4")
         self.port.write(cmd)
+        self._lastcommand = cmd
 
         if self.stabilizationDelay and sync:
             time.sleep(self.stabilizationDelay)
@@ -624,6 +652,7 @@ class ElectronGunControl:
 
         cmd = cmd + b'\n'
         self.port.write(cmd)
+        self._lastcommand = cmd
 
         if self.stabilizationDelay and sync:
             time.sleep(self.stabilizationDelay)
@@ -641,6 +670,7 @@ class ElectronGunControl:
 
         cmd = b'$$$setfila' + bytes(str(currentMa), encoding="ascii") + b'\n'
         self.port.write(cmd)
+        self._lastcommand = cmd
 
         if self.stabilizationDelay and sync:
             time.sleep(self.stabilizationDelay)
@@ -651,6 +681,7 @@ class ElectronGunControl:
 
         cmd = b'$$$filon\n'
         self.port.write(cmd)
+        self._lastcommand = cmd
 
         if self.stabilizationDelay and sync:
             time.sleep(self.stabilizationDelay)
@@ -661,6 +692,7 @@ class ElectronGunControl:
 
         cmd = b'$$$filoff\n'
         self.port.write(cmd)
+        self._lastcommand = cmd
 
         if self.stabilizationDelay and sync:
             time.sleep(self.stabilizationDelay)
@@ -669,6 +701,7 @@ class ElectronGunControl:
         if self.port == False:
             raise ElectronGunNotConnected("Electron gun currently not connected")
         self.port.write(b'$$$insul\n')
+        self._lastcommand = b'$$$insul\n'
         if sync:
             return self.internal__waitForMessageFilter("insulok")
         else:
@@ -678,6 +711,7 @@ class ElectronGunControl:
         if self.port == False:
             raise ElectronGunNotConnected("Electron gun currently not connected")
         self.port.write(b'$$$beamon\n')
+        self._lastcommand = b'$$$beamon\n'
         if sync:
             return self.internal__waitForMessageFilter("beamon")
         else:
@@ -687,9 +721,11 @@ class ElectronGunControl:
         if self.port == False:
             raise ElectronGunNotConnected("Electron gun currently not connected")
         self.port.write(b'$$$reset\n')
+        self._lastcommand = b'$$$reset\n'
         if sync:
             time.sleep(10)
             self.port.write(b'$$$id\n')
+            self._lastcommand = b'$$$id\n'
             if sync:
                 return self.internal__waitForMessageFilter("id")
             else:
@@ -703,3 +739,279 @@ class ElectronGunControl:
             return self.internal__waitForMessageFilter("id")
         else:
             return None
+
+# MQTT implementation
+
+import paho.mqtt.client as mqtt
+import mqttpattern
+
+class ElectronGunControlMQTT:
+    def __init__(self, broker=None, port=1883, user=None, password=None, basetopic=None, Sync = True):
+        if not broker:
+            try:
+                from quakesrctrl import config
+                if config.MQTT_EBEAM:
+                    broker = config.MQTT_EBEAM['broker']
+                    port = config.MQTT_EBEAM['port']
+                    user = config.MQTT_EBEAM['user']
+                    password = config.MQTT_EBEAM['password']
+                    basetopic = config.MQTT_EBEAM['basetopic']
+            except Exception:
+                pass
+
+        if (not broker) or (not basetopic):
+            raise ElectronGunInvalidParameterException("At least broker and basetopic have to be specified")
+
+        if basetopic[-1] != '/':
+            basetopic = basetopic + '/'
+        self._basetopic = basetopic
+
+        self.messageFilter = None
+        self.messageResponse = None
+        self.messageConditionVariable = threading.Condition()
+
+        self.cbIdentify = None
+        self.cbVoltage = None
+        self.cbCurrent = None
+        self.cbPSUMode = None
+        self.cbFilamentCurrent = None
+        self.cbInsulation = None
+        self.cbBeamon = None
+        self.cbFilamentCurrentSet = None
+        self.cbOff = None
+
+        self.mqttPattern = mqttpattern.MQTTPatternMatcher()
+        self.mqttPattern.registerHandler(f"{self._basetopic}egun/id", self._msghandler_Id)
+        self.mqttPattern.registerHandler(f"{self._basetopic}egun/voltage", self._msghandler_Voltage)
+        self.mqttPattern.registerHandler(f"{self._basetopic}egun/current", self._msghandler_Current)
+
+        self.mqtt = mqtt.Client()
+        self.mqtt.on_connect = self._mqtt_on_connect
+        self.mqtt.on_message = self._mqtt_on_message
+        if user:
+            self.mqtt.username_pw_set(user, password)
+        self.mqtt.connect(broker, port)
+        self.mqtt.loop_start()
+
+        while Sync:
+            if self.internal__waitForMessageFilter("connect"):
+                break
+
+        atexit.register(self.close)
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        atexit.unregister(self.close)
+        if self.mqtt:
+            self.mqtt.disconnect()
+            self.mqtt = None
+
+    def close():
+        atexit.unregister(self.close)
+        if self.mqtt:
+            self.mqtt.disconnect()
+            self.mqtt = None
+
+    def internal__waitForMessageFilter(self, filter):
+        self.messageConditionVariable.acquire()
+
+        self.messageFilter = filter
+
+        # Now wait till we get a response from our processing thread ...
+        while self.messageResponse == None:
+            self.messageConditionVariable.wait()
+        # Reset message filter, copy and release response ...
+        self.messageFilter = None
+        retval = self.messageResponse
+        self.messageResponse = None
+
+        # Release ...
+        self.messageConditionVariable.release()
+        return retval
+
+    def internal__signalCondition(self, messageType, payload):
+        self.messageConditionVariable.acquire()
+        if messageType != self.messageFilter:
+            self.messageConditionVariable.release()
+            return
+        self.messageResponse = payload
+        self.messageConditionVariable.notify_all()
+        self.messageConditionVariable.release()
+
+    def id(self, *ignore, sync = False):
+        if not self.mqtt:
+            raise ElectronGunNotConnected("Electron gun currently not connected")
+        self._mqtt_publish("egun/id/request")
+        if sync:
+            return self.internal__waitForMessageFilter("id")
+        else:
+            return None
+
+    def _msghandler_Id(self, message):
+        if self.cbIdentify:
+            if type(self.cbIdentify) is list:
+                for f in self.cbIdentify:
+                    if callable(f):
+                        f(self, message.payload['version'], message.payload['revision'])
+            elif callable(self.cbIdentify):
+                self.cbIdentify(self, message.payload['version'], message.payload['revision'])
+
+        self.internal__signalCondition("id", { 'version' : message.payload['version'], 'revision' : message.payload['revision'] })
+
+    def getPSUVoltage(self, channel, *ignore, sync = False):
+        if not self.mqtt:
+            raise ElectronGunNotConnected("Electron gun currently not connected")
+        if (channel < 1) or (channel > 4):
+            raise ElectronGunInvalidParameterException("Power supply channel has to be in range 1 to 4")
+
+        self._mqtt_publish(f"egun/voltage/request", { 'channel' : channel })
+
+        if sync:
+            res = self.internal__waitForMessageFilter("v{}".format(channel))
+            return res
+        else:
+            return None
+
+    def _msghandler_Voltage(self, message):
+        channel = message.payload['channel']
+        voltage = message.payload['voltage']
+
+        if self.cbVoltage:
+            if type(self.cbVoltage) is list:
+                for f in self.cbVoltage:
+                    if callable(f):
+                        f(self, channel, voltage)
+            elif callable(self.cbVoltage):
+                self.cbVoltage(self, channel, voltage)
+        self.internal__signalCondition("v{}".format(channel), voltage )
+
+    def quakEstimateBeamCurrent(self):
+        if self.port == False:
+            raise ElectronGunNotConnected("Electron gun currently not connected")
+
+        # Power supplies:
+        #   1   Cathode
+        #   2   Whenelt
+        #   3   Focus
+        #   4   unused
+        currentCathode = self.getPSUCurrent(1, sync = True)
+        currentWhenelt = self.getPSUCurrent(2, sync = True)
+        currentFocus = self.getPSUCurrent(3, sync = True)
+
+        return currentCathode + currentWhenelt + currentFocus
+
+    def getPSUCurrent(self, channel, *ignore, sync = False):
+        if not self.mqtt:
+            raise ElectronGunNotConnected("Electron gun currently not connected")
+        if (channel < 1) or (channel > 4):
+            raise ElectronGunInvalidParameterException("Power supply channel has to be in range 1 to 4")
+
+        self._mqtt_publish(f"egun/current/request", { 'channel' : channel })
+
+        if sync:
+            res = self.internal__waitForMessageFilter("a{}".format(channel))
+            return res
+        else:
+            return None
+
+    def _msghandler_Current(self, message):
+        channel = message.payload['channel']
+        current = message.payload['current']
+
+        if self.cbCurrent:
+            if type(self.cbCurrent) is list:
+                for f in self.cbCurrent:
+                    if callable(f):
+                        f(self, channel, current)
+            elif callable(self.cbCurrent):
+                self.cbVoltage(self, channel, current)
+        self.internal__signalCondition("a{}".format(channel), current)
+
+    def getPSUModes(self, *ignore, sync = False):
+        pass
+
+    def getFilamentCurrent(self, *ignore, sync = False):
+        pass
+
+    def off(self, *ignore, sync = False):
+        pass
+
+    def noprotection(self, *ignore, sync = False):
+        pass
+
+    def setPSUPolarity(self, channel, polarity, *ignore, sync = False):
+        pass
+
+    def setPSUEnable(self, channel, *ignore, sync = False):
+        pass
+
+    def setPSUDisable(self, channel, *ignore, sync = False):
+        pass
+
+    def setPSUVoltage(self, channel, voltage, *ignore, sync = False):
+        pass
+
+    def setFilamentCurrent(self, currentMa, *ignore, sync = False):
+        if not self.mqtt:
+            raise ElectronGunNotConnected("Electron gun currently not connected")
+        if (currentMa < 0) or (currentMa > 200):
+            raise ValueError("Current is out of range")
+
+        self._mqtt_publish(f"egun/filamentcurrent/set", { 'current' : currentMa })
+
+        if sync:
+            res = self.internal__waitForMessageFilter("a{}".format(channel))
+            return res
+        else:
+            return None
+
+    def setFilamentOn(self, *ignore, sync = False):
+        pass
+
+    def setFilamentOff(self, *ignore, sync = False):
+        pass
+
+    def runInsulationTest(self, *ignore, sync = False):
+        pass
+
+    def beamOn(self, *ignore, sync = False):
+        pass
+
+    def reset(self, *ignore, sync = False):
+        pass
+
+    def _mqtt_publish(self, topic, message=None, prependBaseTopic=True, retain=False):
+        if not self.mqtt:
+            raise ElectronGunNotConnected("Electron gun currently not connected")
+
+        if isinstance(message, dict):
+            message = json.dumps(message)
+
+        if prependBaseTopic:
+            topic = self._basetopic + topic
+
+        try:
+            if not (message is None):
+                self.mqtt.publish(topic, payload=message, qos=0, retain=retain)
+            else:
+                self.mqtt.publish(topic, qos=0, retain=retain)
+            return True
+        except Exception as e:
+            return False
+
+    def _mqtt_on_connect(self, client, userdata, flags, rc):
+        if rc == 0:
+            self.internal__signalCondition("connect", { 'success' : True })
+            client.subscribe(self._basetopic+"#")
+        else:
+            self.internal__signalCondition("connect", { 'success' : False })
+
+    def _mqtt_on_message(self, client, userdata, msg):
+        try:
+            msg.payload = json.loads(str(msg.payload.decode('utf-8', 'ignore')))
+        except:
+            # Ignore if we don't have a JSON payload
+            pass
+        self.mqttPattern.callHandlers(msg.topic, msg)
